"""
Description
-----------
This small Python package allows to simply draw upper air sounding on a Skew-T diagramm using
``metpy`` library.

License
-------
All this project is under GNU General Public License v3.0+ (GPLv3+).
Please see the LICENSE file for further informations.

Links
-----
* Home page: https://github.com/Shadow15510/cra
* Documentation: http://cra.rtfd.io/
"""

__version__ = "1.1.4"



"""
This module provide a class to manipulate variables with dimensions.

Classes
-------
.. autoclass:: LockedWarning
.. autoclass:: VariablesSet
"""

from warnings import warn
import numpy as np
import pandas as pd
import metpy.calc as mpcalc


class LockedWarning(Warning):
    """Warns the user that he's trying to modify a locked VariablesSet instance."""


class VariablesSet:
    """
    This class provides a standard format for passing variables to an AirSounding instance. It also
    allows data to be split between ascending and descending profiles.

    Attributes
    ----------
    data : pd.DataFrame, private
        The DataFrame that contains all the data from the upper air sounding.
    variables : dict, private
        Identified variables in ``data``. You can only identify the variables required by
        AirSounding, in particular :

        * the pressure;

        * the temperature;

        * the dewpoint;

        * the easterly wind;

        * the northerly wind.

    locked : bool, private
        ``False`` if the instance can still be edited.

    Methods
    -------
    .. automethod:: add_press
    .. automethod:: add_temp
    .. automethod:: add_dewpoint
    .. automethod:: add_windu
    .. automethod:: add_windv
    .. automethod:: compute_dewpoint
    .. automethod:: apply_threshold
    """

    def __init__(self, filename: str, **kwargs):
        """Constructor method."""
        self.__data = pd.read_csv(filename, **kwargs)
        self.__data.dropna(inplace=True)
        self.__variables = {}
        self.__locked = False

    def __getitem__(self, varname: str):
        """
        Returns the called variable

        Parameters
        ----------
        varname : str
            The name of the requested variable.
        """
        return self.__variables[varname]

    def add_press(self, vardata: tuple):
        """
        Adds pressure to ``variables``.

        Parameters
        ----------
        vardata : tuple
            This tuple must contains:

            * the name of the pressure in ``data``;

            * the unit of the pressure in ``data`` (cf. ``metpy.units``).

        Warns
        -----
        LockedWarning
            This warning should appear if you try to add a pressure when the VariablesSet is locked.
        """
        if not self.__locked:
            self.__variables["press"] = self.__data[vardata[0]].values * vardata[1]
        else:
            warn(
                "this instance of VariableSet is locked, you cannot add a new variable to it",
                LockedWarning,
            )

    def add_temp(self, vardata: tuple):
        """
        Adds temperature to ``variables``.

        Parameters
        ----------
        vardata : tuple
            This tuple must contains:

            * the name of the temperature in ``data``;

            * the unit of the temperature in ``data`` (cf. ``metpy.units``).

        Warns
        -----
        LockedWarning
            This warning should appear if you try to add a temperature when the VariablesSet is
            locked.
        """
        if not self.__locked:
            self.__variables["temp"] = self.__data[vardata[0]].values * vardata[1]
        else:
            warn(
                "this instance of VariableSet is locked, you cannot add a new variable to it",
                LockedWarning,
            )

    def add_dewpoint(self, vardata: tuple):
        """
        Adds dewpoint to ``variables``.

        Parameters
        ----------
        vardata : tuple
            This tuple must contains:

            * the name of the dewpoint in ``data``;

            * the unit of the dewpoint in ``data`` (cf. ``metpy.units``).

        Warns
        -----
        LockedWarning
            This warning should appear if you try to add a dewpoint when the VariablesSet is locked.
        """
        if not self.__locked:
            self.__variables["dewpoint"] = self.__data[vardata[0]].values * vardata[1]
        else:
            warn(
                "this instance of VariableSet is locked, you cannot add a new variable to it",
                LockedWarning,
            )

    def add_windu(self, vardata: tuple):
        """
        Adds east wind component to ``variables``.

        Parameters
        ----------
        vardata : tuple
            This tuple must contains:

            * the name of the east wind component in ``data``;

            * the unit of the east wind component in ``data`` (cf. ``metpy.units``).

        Warns
        -----
        LockedWarning
            This warning should appear if you try to add an east component of the wind when the
            VariablesSet is locked.
        """
        if not self.__locked:
            self.__variables["wind_u"] = self.__data[vardata[0]].values * vardata[1]
        else:
            warn(
                "this instance of VariableSet is locked, you cannot add a new variable to it",
                LockedWarning,
            )

    def add_windv(self, vardata: tuple):
        """
        Adds north wind component to ``variables``.

        Parameters
        ----------
        vardata : tuple
            This tuple must contains:

            * the name of the north wind component in ``data``;

            * the unit of the north wind component in ``data`` (cf. ``metpy.units``).

        Warns
        -----
        LockedWarning
            This warning should appear if you try to add a north component of the wind when the
            VariablesSet is locked.
        """
        if not self.__locked:
            self.__variables["wind_v"] = self.__data[vardata[0]].values * vardata[1]
        else:
            warn(
                "this instance of VariableSet is locked, you cannot add a new variable to it",
                LockedWarning,
            )

    def compute_dewpoint(self, temperature: tuple, relative_humitidy: tuple):
        """
        Computes the dewpoint from the air temperature and the relative humidity.

        Parameters
        ----------
        temperature : tuple
            This tuple must contains:

            * the name of the temperature in ``data``;

            * the unit of the temperature in ``data`` (cf. ``metpy.units``).

        relative_humidity : tuple
            This tuple must contains:

            * the name of the relative humidity in ``data``;

            * the unit of the relative humidity in ``data`` (cf. ``metpy.units``).

        Warns
        -----
        LockedWarning
            This warning should appear if you try to compute the dewpoint when the
            VariablesSet is locked.
        """
        if "dewpoint" in self.__variables:
            raise KeyError("there's already an entry for the dewpoint")

        if not self.__locked:
            self.__variables["dewpoint"] = mpcalc.dewpoint_from_relative_humidity(
                self.__data[temperature[0]].values * temperature[1],
                self.__data[relative_humitidy[0]].values * relative_humitidy[1],
            ).to("degC")
        else:
            warn(
                "this instance of VariableSet is locked, you cannot add a new variable to it",
                LockedWarning,
            )

    def apply_threshold(self, threshold: float, ascending: bool = True, start_cut: bool = False,
        atol: int = 10
    ):
        """
        This function cuts data by applying a threshold to the pressure. It will retain all data
        above this threshold. You can also request an ascending or descending profile for altitude
        sounding.

        .. warning::
            Applying a threshold will lock the ``VariablesSet`` instance.

        Parameters
        ----------
        threshold : float
            The threshold to apply on pressure, all pressure above this value will be retain.
        ascending : bool, optionnal
            By default: ``True``.
            Indicates which profile you want to keep. If ``ascending`` is set on ``True`` it will
            keep the ascending profile, otherwise the descending profile.
        start_cut : bool, optionnal
            By default: ``False``
            If set on ``True`` it will cuts the data at the start.
        atol : int, optionnal
            By default: ``10``
            The absolute tolerance parameter.

        Warns
        -----
        LockedWarning
            This warning should appear if you try to apply a threshold when the
            VariablesSet is locked.
        """
        if not self.__locked:
            threshold = np.where(
                np.isclose(self.__variables["press"].magnitude, threshold, atol=atol)
            )[0]

            if not start_cut:
                self.__locked = True

            if ascending:
                threshold = threshold[0]
                for var in self.__variables:
                    if start_cut:
                        self.__variables[var] = self.__variables[var][threshold: ]
                    else:
                        self.__variables[var] = self.__variables[var][:threshold]
            else:
                threshold = threshold[1]
                for var in self.__variables:
                    self.__variables[var] = np.flip(self.__variables[var][threshold:])

        else:
            warn(
                "this instance of VariableSet is locked, you cannot add a new variable to it",
                LockedWarning,
            )


"""
This module provides functions for dividing a signal into layers by analyzing the strongest
gradients into the input signal.

Functions
---------
.. autofunction:: check_indices_spacing
.. autofunction:: get_layers
"""

import numpy as np


def check_indices_spacing(indices: list, index_to_check: int, min_dist: int):
    """
    Checks that the index to be checked is far enough away from all the others.

    Parameters
    ----------
    indices : list
        All the indices, they're considered as a reference.
    index_to_check : int
        The index to be checked.
    min_dist : int
        The minimum distance beetween two indices.

    Returns
    -------
    out : bool
        It equals ``True`` if ``index_to_check`` is far enough away from all the others, and
        ``False`` in the other case.
    """
    for i in indices:
        if abs(index_to_check - i) <= min_dist:
            return False
    return True


def get_layers(layer_size: int, nb_layers: int, *datas):
    """
    Extracts strongest gradients from input data. If you give several input datas, the result will
    be the concatenation of all the gradients found.

    Parameters
    ----------
    layer_size : int
        The minimum size of each layer (in terms of index). The input datas will be divided into
        small layers of this size, and the sliding gradients will be computed on these layers.
    nb_layers : int
        The numbers of layers you want.
    datas
        All the data on which you wish to obtain the strongest gradients.

    Returns
    -------
    out : list
        The sorted list of the indices of all the strongest gradients.

    Exemple
    -------
    If you're searching for small phenomena into an upper air sounding, you should pass a small
    ``layer_size``. You also have two variables from the sounding (e.g. temperature and humidty)
    and you want 10 layers::

        layers_indices = get_layers(10, 10, data_temp, data_relative_hum)
    """
    final_indices = set()
    nb_layers //= len(datas)

    for data in datas:
        nb_data = len(data)
        gradients = []
        for i in range(nb_data - layer_size):
            grads = np.abs(np.gradient(data[i: i + layer_size]))
            gradients.append((i, sum(grads) / layer_size))

        indices = [0]
        while len(indices) < nb_layers and gradients:
            grad_max = max(gradients, key=lambda x: x[1])
            gradients.remove(grad_max)
            if check_indices_spacing(indices, grad_max[0], layer_size):
                indices.append(grad_max[0])
                indices.append(grad_max[0] + layer_size)
        indices.append(nb_data - 1)

        final_indices = {*final_indices, *indices}

    return sorted(list(final_indices))



"""
This module allows to plot an upper air sounding on a Skew-T diagramm. It can also try to detect
interesting layers into the upper air sounding.

Classes
-------
.. autoclass:: AirSounding
"""

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import metpy.calc as mpcalc
from metpy.units import units
from metpy.plots import SkewT, Hodograph
from cra.layers import get_layers
from cra.variables import VariablesSet


# LaTeX-like font setup
matplotlib.rcParams["mathtext.fontset"] = "stix"
matplotlib.rcParams["font.family"] = "STIXGeneral"


class AirSounding:
    """
    Allows to plot and analyse an upper air sounding

    Attributes
    ----------
    nb_data : int
        The length of the given ``VariablesSet``.
    data : dict
        The variables contained into the given ``VariablesSet``.

    Methods
    -------
    .. automethod:: show
    .. automethod:: auto_layers
    """

    def __init__(self, variables: VariablesSet):
        """Constructor method."""
        self.nb_data = len(variables["press"])
        self.data = {
            "press": variables["press"].to("hPa"),
            "temp": variables["temp"].to("degC"),
            "p_temp": np.zeros(self.nb_data) * units.degC,
            "wbp_temp": np.zeros(self.nb_data) * units.degC,
            "wb_temp": np.zeros(self.nb_data) * units.degC,
            "dewpoint": variables["dewpoint"].to("degC"),
            "wind_u": variables["wind_u"].to("m/s"),
            "wind_v": variables["wind_v"].to("m/s"),
        }

        for index in range(self.nb_data):
            self.data["p_temp"][index] = mpcalc.potential_temperature(
                self.data["press"][index], self.data["temp"][index]
            ).to("degC")
            self.data["wbp_temp"][index] = mpcalc.wet_bulb_potential_temperature(
                self.data["press"][index],
                self.data["temp"][index],
                self.data["dewpoint"][index],
            ).to("degC")
            self.data["wb_temp"][index] = mpcalc.wet_bulb_temperature(
                self.data["press"][index],
                self.data["temp"][index],
                self.data["dewpoint"][index],
            ).to("degC")

    def __layers_analysis(self, layers: list):
        """
        Analyses the found layers of the upper air sounding.

        Parameters
        ----------
        layers : list
            The indexes of the layers.

        Returns
        -------
        out : tuple, yield
            The returned tuple contains four pieces of information on the analysed layer:

            * the evolution of the potential temperature;

            * the evolution of the wet bulb potential temperature;

            * the saturation at the bottom of the layer;

            * the saturation at the top of the layer.
        """
        for index in range(len(layers) - 1):
            pt_start = mpcalc.potential_temperature(
                self.data["press"][index], self.data["temp"][index]
            )
            pt_end = mpcalc.potential_temperature(
                self.data["press"][index + 1], self.data["temp"][index + 1]
            )

            wbpt_start = mpcalc.wet_bulb_potential_temperature(
                self.data["press"][index],
                self.data["temp"][index],
                self.data["dewpoint"][index],
            )
            wbpt_end = mpcalc.wet_bulb_potential_temperature(
                self.data["press"][index + 1],
                self.data["temp"][index + 1],
                self.data["dewpoint"][index + 1],
            )

            yield (
                (pt_end - pt_start) > 0,
                (wbpt_end - wbpt_start) > 0,
                (
                    self.data["temp"][index].magnitude
                    - self.data["dewpoint"][index].magnitude
                )
                < 0.2,
                (
                    self.data["temp"][index + 1].magnitude
                    - self.data["dewpoint"][index + 1].magnitude
                )
                < 0.2,
            )

    def show(self, layers: tuple = ()):
        """
        Plots and displays the upper air sounding on a Skew-T diagramm and a hodograph.

        Parameters
        ----------
        layers : tuple, optionnal
            By default: ``()``.
            The layers to be represented on the Skew-T diagramm.
        """
        fig = plt.figure(figsize=(18, 10))
        idx = mpcalc.resample_nn_1d(
            self.data["press"], np.arange(100, 1000, 25) * units.hPa
        )
        profile = mpcalc.parcel_profile(
            self.data["press"], self.data["temp"][0], self.data["dewpoint"][0]
        ).to("degC")

        skew = SkewT(fig, rect=(0.05, 0.05, 0.5, 0.90))
        skew.plot(self.data["press"], self.data["temp"], "r", lw=2, label="Température")
        skew.plot(
            self.data["press"],
            self.data["wb_temp"],
            "b",
            lw=2,
            label="Température du thermomètre mouillé",
        )
        skew.plot_barbs(
            pressure=self.data["press"][idx],
            u=self.data["wind_u"][idx],
            v=self.data["wind_v"][idx],
        )
        skew.plot(self.data["press"], profile, "k", lw=1)
        skew.shade_cin(
            self.data["press"],
            self.data["temp"],
            profile,
            self.data["dewpoint"],
            color="blue",
            alpha=0.3,
            label="CIN",
        )
        skew.shade_cape(
            self.data["press"],
            self.data["temp"],
            profile,
            color="red",
            alpha=0.3,
            label="CAPE",
        )
        if layers:
            skew.plot(
                layers[0],
                layers[1],
                "o",
                color="darkorange",
                label="Couches auto-détectées",
            )
        skew.plot_dry_adiabats(ls="solid", colors="forestgreen", alpha=0.75)
        skew.plot_moist_adiabats(ls="dashed", colors="forestgreen", alpha=0.75)
        skew.ax.set_xlabel("Température (°C)")
        skew.ax.set_ylabel("Pression (hPa)")

        hodo = Hodograph(plt.axes((0.45, 0.51, 0.44, 0.44)), component_range=50)
        hodo.add_grid(increment=10, ls="-", lw=1.5, alpha=0.75)
        hodo.add_grid(increment=5, ls="--", lw=1, alpha=0.75)
        hodo.plot_colormapped(
            self.data["wind_u"], self.data["wind_v"], c=self.data["press"], label="Vent"
        )
        hodo.ax.set_box_aspect(1)
        hodo.ax.set_xticks([])
        hodo.ax.set_yticks([])
        plt.xticks(np.arange(0, 0, 1))
        plt.yticks(np.arange(0, 0, 1))
        for i in range(10, 120, 10):
            hodo.ax.annotate(
                str(i),
                (i, 0),
                xytext=(0, 2),
                textcoords="offset pixels",
                clip_on=True,
                fontsize=10,
                weight="bold",
                alpha=0.3,
                zorder=0,
            )
            hodo.ax.annotate(
                str(i),
                (0, i),
                xytext=(0, 2),
                textcoords="offset pixels",
                clip_on=True,
                fontsize=10,
                weight="bold",
                alpha=0.3,
                zorder=0,
            )

        # hodo.ax.set_xlabel(r'Vitesse (Est, m$\cdot$s$^{-1}$)')
        # hodo.ax.set_ylabel(r'Vitesse (Nord, m$\cdot$s$^{-1}$)')

        # fig.patches.extend([plt.Rectangle(
        #         (0.56, 0.05), 0.39, 0.44,
        #         edgecolor='black',
        #         facecolor='white',
        #         lw=1,
        #         alpha=1,
        #         transform=fig.transFigure,
        #         figure=fig
        #     )])
        # # espace de 0.025
        # try:
        #     cape, cin = mpcalc.surface_based_cape_cin(
        #             self.data['press'],
        #             self.data['temp'],
        #             self.data['dewpoint']
        #         )
        #     cape = round(cape.magnitude, 0)
        #     cin = round(cin.magnitude, 0)
        # except ValueError:
        #     cape = cin = np.nan
        # plt.figtext(0.562, 0.47, 'Informations', fontsize=15, weight='bold')
        # plt.figtext(0.562, 0.445, rf'CAPE : {cape} J$\cdot$kg$^{{-1}}$', fontsize=15)
        # plt.figtext(0.562, 0.42, rf'CIN : {cin} J$\cdot$kg$^{{-1}}$', fontsize=15)

        skew.ax.legend(loc="upper left")
        hodo.ax.legend(loc="upper left")
        plt.show()

    def auto_layers(self, layer_size: int = 0, nb_layers: int = 10, show: bool = False):
        """
        This method attempts to find interesting layers in the upper air sounding and prints the
        characteristics of these layers.

        Parameters
        ----------
        layer_size : int, optionnal
            By default: automatic size.
            The size of each layers in termes of indexes.
        nb_layers : int, optionnal
            By default: ``10``.
            The maximum number of layers you want.
        show : bool, optionnal
            By default: ``False``.
            If set on ``True``, once the layers have been analyzed, it will plot the layers on the
            Skew-T diagramm.
        """

        def get_evolution(delta: float):
            if delta > 0:
                return "increase"
            return "decrease"

        if not layer_size:
            layer_size = self.nb_data // nb_layers

        layers = get_layers(
            layer_size,
            nb_layers,
            self.data["temp"],
            self.data["dewpoint"],
            self.data["p_temp"],
            self.data["wbp_temp"],
        )
        press = self.data["press"][layers].magnitude
        temp = self.data["temp"][layers].magnitude
        for index, info in enumerate(self.__layers_analysis(layers)):
            title = f"Layer n°{index + 1}"
            print(f"{title}\n" + len(title) * "-")

            print("Limits")
            print(f".. pressure    : {press[index]} hPa – {press[index + 1]} hPa")
            print(f".. temperature : {temp[index]}°C – {temp[index + 1]}°C")

            print("Information")
            print(f".. potential temperature          : {get_evolution(info[0])}")
            print(f".. wet bulb potential temperature : {get_evolution(info[1])}")
            print(f".. saturation base                : {info[2]}")
            print(f".. saturation top                 : {info[3]}")

            print()

        if show:
            self.show((press, temp))



"""
This little module gives some functions to handle common file formats.

Functions
---------
.. autofunction:: open_ftr
.. autofunction:: open_cor
"""

from metpy.units import units
from cra.variables import VariablesSet


def open_ftr(filename: str, **kwargs):
    """
    Opens and creates a ``VariablesSet`` instance from a ftr file.

    Parameters
    ----------
    filename : str
        The name of the ftr file to be read.
    kwargs
        The keyword arguments to be given to ``pandas.read_csv``.

    Returns
    -------
    v_set : VariablesSet
        The ``VariablesSet`` instance, ready to be given to ``AirSounding``.
    """
    v_set = VariablesSet(filename, **kwargs)
    v_set.add_press(("ftr_pres", units.hPa))
    v_set.add_temp(("ftr_temp", units.degC))
    v_set.add_dewpoint(("ftr_DP", units.degC))
    v_set.add_windu(("ftr_VEF", units("m/s")))
    v_set.add_windv(("ftr_VNF", units("m/s")))
    return v_set


def open_cor(filename: str, **kwargs):
    """
    Opens and creates a ``VariablesSet`` instance from a cor file.

    Parameters
    ----------
    filename : str
        The name of the cor file to be read.
    kwargs
        The keyword arguments to be given to ``pandas.read_csv``.

    Returns
    -------
    v_set : VariablesSet
        The ``VariablesSet`` instance, ready to be given to ``AirSounding``.
    """
    v_set = VariablesSet(filename, **kwargs)
    v_set.add_press(("Press", units.hPa))
    v_set.add_temp(("TaRad", units.degC))
    v_set.compute_dewpoint(("TaRad", units.degC), ("UCal", units.percent))
    v_set.add_windu(("VE", units("m/s")))
    v_set.add_windv(("VN", units("m/s")))
    return v_set
